#include "udm_config.h"

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_IO_H   /* for Win */
#include <io.h>
#endif
#ifdef HAVE_PROCESS_H
#include <process.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_MMAN_H
#include <sys/mman.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif


#include "udm_common.h"
#include "udm_cache.h"
#include "udm_utils.h"
#include "udm_db.h"

#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

/*#define DEBUG_LOGD 1*/
/*#define DEBUG_LOGD_CMD 1*/

#define SPLDIR  "splitter"

#define STATE_CMD 1
#define STATE_WRD 2
#define MAXCLIENT 128

/************************ type defenitions ************************/ 
typedef struct {
	int  fd;	/* Socket descriptor    */
	size_t  nbytes;	/* Bytes in buffer      */
	size_t  rbytes;	/* Required bytes       */
	size_t  mbytes;	/* Allocated bytes      */
	int  state;	/* Client state         */
	UDM_LOGD_CMD last_cmd;
	union {
		char * buf;	/* Buffer for data   */
		UDM_LOGD_CMD * cmd;
		UDM_LOGD_WRD * wrd;
	} data;	
} UDM_LOGD_CL;

#ifdef DEBUG_LOGD_CMD
void printcmd(int id,UDM_LOGD_CMD *c){
	fprintf(stderr,"cmd=%d nwords=%d stamp=%d url_id=%d\n",c->cmd,c->nwords,(int)c->stamp,c->url_id);
}
#endif

static void init_client(UDM_LOGD_CL * client, int n){
	int i;
	for(i=0;i<n;i++){
		client[i].fd=0;
		client[i].state=0;
		client[i].nbytes=0;
		client[i].rbytes=0;
		client[i].mbytes=1;
		client[i].data.buf=(char*)malloc(1);
	}
}

/********************* SIG Handlers ****************/
static void sighandler(int sign);

static void init_signals(void){
	/* Set up signals handler*/
#if (WIN32|WINNT)
#else
	signal(SIGPIPE,sighandler);
	signal(SIGHUP,sighandler);
#endif
	signal(SIGINT,sighandler);
	signal(SIGTERM,sighandler);
}

static void sighandler(int sign){

	fprintf(stderr,"%s Signal %d arrived\n",Logd_time_pid_info(&LOGD),sign);
	switch(sign){
#if (WIN32|WINNT)
#else
		case SIGPIPE:
			fprintf(stderr,"%s FATAL: This should not ever happen! Broken pipe!\n",Logd_time_pid_info(&LOGD));
			break;
		case SIGHUP:
			LogdSaveAllBufs(&LOGD);
			break;
#endif
		default:
			fprintf(stderr,"%s Shutdown\n",Logd_time_pid_info(&LOGD));
			exit(0);
	}
	init_signals();
}

/*************************************************************/

static int client_action(int id,UDM_LOGD_CL * client){

	switch(client->state){
		
		case STATE_CMD:
			client->last_cmd=*(client->data.cmd);
#ifdef DEBUG_LOGD_CMD
			printcmd(id,client->data.cmd);
#endif
			if(client->last_cmd.nwords>0){
				client->state=STATE_WRD;
				client->rbytes=client->last_cmd.nwords*sizeof(UDM_LOGD_WRD);
			}else{
				if(LogdStoreDoc(&LOGD,client->last_cmd,client->data.wrd)) exit(1);
				client->state=STATE_CMD;
				client->rbytes=sizeof(UDM_LOGD_CMD);
			}
			break;
		case STATE_WRD:
			/*
			switch(fork()){
				case 0:	if(LogdStoreDoc(client->last_cmd,client->data.wrd)) exit(1);
					exit(0);
					break;
				case -1:fprintf(stderr,"%s Can't fork: '%s'\n",Logd_time_pid_info(&LOGD),strerror(errno));
					break;
			}
			*/
			if(LogdStoreDoc(&LOGD,client->last_cmd,client->data.wrd)) exit(1);
			client->state=STATE_CMD;
			client->rbytes=sizeof(UDM_LOGD_CMD);
			break;
	}
	return 0;
}

static void usage(void){

	fprintf(stderr,
"\n\
cachelogd from %s-%s-%s\n\
http://search.mnogo.ru (C) 1998-2001, LavTech Corp.\n\
\n\
Usage: cachelogd [OPTIONS]\n\
\n\
Options are:\n\
  -w /path      choose alternative working /var directory\n\
  -p xxx        listen port xxx\n\
  -h,-?         print this help page and exit\n\
\n\
Please mail bug reports and suggestions to <devel@mnogosearch.org>.\n",
	PACKAGE,VERSION,UDM_DBTYPE);

	return;
}

static char pid_name[1024];

static void exitproc(void){
	LogdClose(&LOGD);
	unlink(pid_name);
#if (WIN32|WINNT)
	WSACleanup();
#endif
}

int main(int argc,char **argv){
	int flags=O_RDWR|O_CREAT|O_APPEND|UDM_BINARY;
	int perm=UDM_IWRITE;
	struct	sockaddr_in server_addr;
	struct	sockaddr_in his_addr;
	struct	in_addr bind_address;
	
	char var_dir[1024]=UDM_VAR_DIR UDMSLASHSTR;

	int on=1,i;
	int ctl_sock, sel;
	struct timeval tval;

	int  pid_fd,ch,port=UDM_LOGD_PORT;
	char pidbuf[1024];
	fd_set mask;
	UDM_LOGD_CL log_client[MAXCLIENT];
	
	while ((ch = getopt(argc, argv, "w:p:?")) != -1){
		switch (ch) {
			case 'p':
				port=atoi(optarg);
				break;
			case 'w':
				strcpy(var_dir,optarg);
				break;
			case 'h':
			case '?':
			default:
				usage();
				return 1;
				break;
		}
	}
	argc -= optind;argv += optind;

	if(argc>0){
		usage();
		return 1;
	}

	/* Check that another instance isn't running and create PID file */
	sprintf(pid_name,"%s%s",var_dir,"cachelogd.pid");
	pid_fd=open(pid_name,flags,perm);
	if(pid_fd<0){
		fprintf(stderr,"%s Can't create '%s': %s\n",Logd_time_pid_info(&LOGD),pid_name,strerror(errno));
		if(errno==EEXIST){
			fprintf(stderr,"It seems that another cachelogd is already running!\n");
			fprintf(stderr,"Remove '%s' if it is not true.\n",pid_name);
		}
		exit(1);
	}
	sprintf(pidbuf,"%d\n",(int)getpid());
	write(pid_fd,&pidbuf,strlen(pidbuf));
	close(pid_fd);

#if (WIN32|WINNT)
	{
		WSADATA wsaData;

		if(WSAStartup(0x101,&wsaData)!=0){
			fprintf(stderr,"%s WSAStartup() error %d\n",Logd_time_pid_info(&LOGD),WSAGetLastError);
			exit(1);
		}
	}
#endif
	/* Initialize variables */

	LogdInit(&LOGD,var_dir,1);
	atexit(&exitproc);
	init_client(log_client,MAXCLIENT);
	init_signals();

	fprintf(stderr,"%s Started on port %d. Accepting %d connections\n",Logd_time_pid_info(&LOGD),port,sizeof(mask));

	ctl_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (ctl_sock < 0) {
		fprintf(stderr,"%s socket() error %d\n",Logd_time_pid_info(&LOGD),errno);
		unlink(pid_name);
		exit(1);
	}
	if (setsockopt(ctl_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) != 0){
		fprintf(stderr,"%s setsockopt() error %d\n",Logd_time_pid_info(&LOGD),errno);
		unlink(pid_name);
		exit(1);
	}


	/* Prepare to start TCP server */
	
	bind_address.s_addr 	= htonl(INADDR_ANY);
	server_addr.sin_family	= AF_INET;
	server_addr.sin_addr	= bind_address;
	server_addr.sin_port	= ntohs(port);
	
	if (bind(ctl_sock, (struct sockaddr *)&server_addr, sizeof(server_addr))) {
		fprintf(stderr,"%s bind() error %d %s\n",Logd_time_pid_info(&LOGD),errno,strerror(errno));
		unlink(pid_name);
		exit(1);
	}
	if (listen(ctl_sock, 32) < 0) {
		fprintf(stderr,"%s listen() error %d %s\n",Logd_time_pid_info(&LOGD),errno,strerror(errno));
		unlink(pid_name);
		exit(1);
	}

	FD_ZERO(&mask);
	FD_SET(ctl_sock,&mask);

	
	while (1) {
		fd_set msk;
		
		tval.tv_sec  		= 300;
		tval.tv_usec 		= 0;
		msk=mask;
		sel=select(32,&msk,0,0,&tval);
		
		if(sel==0){
			/* Time limit expired */
 			continue;
		}
		if(sel<1){
			/* FIXME: add errno checking */
			continue;
		}
		
		/* Check whether new client has connected */
		if(FD_ISSET(ctl_sock,&msk)){
			int cl,addrlen;
			addrlen = sizeof(his_addr);
			for(cl=0;cl<MAXCLIENT;cl++){
				if(log_client[cl].fd==0){
					fprintf(stderr,"%s Client #%d connected\n",Logd_time_pid_info(&LOGD),cl);
					log_client[cl].fd=accept(ctl_sock, (struct sockaddr *)&his_addr, &addrlen);
					FD_SET(log_client[cl].fd,&mask);
					log_client[cl].state=STATE_CMD;
					log_client[cl].rbytes=sizeof(UDM_LOGD_CMD);
					break;
				}
			}
		}
		
		
		/* Process all ready clients */
		for(i=0;i<MAXCLIENT;i++){
			if(FD_ISSET(log_client[i].fd,&msk)){
				UDM_LOGD_CL * cl;
				int lbytes;
				
				cl=&log_client[i];
#ifdef DEBUG_LOGD
				printf("[%d] try read %d r=%d n=%d\n",i,cl->rbytes-cl->nbytes,cl->rbytes,cl->nbytes);
#endif

				if((cl->mbytes)!=(cl->rbytes)){
					cl->data.buf=(char*)realloc(cl->data.buf,cl->rbytes);
					cl->mbytes=cl->rbytes;
				}

				lbytes=recv(cl->fd,&cl->data.buf[cl->nbytes],cl->rbytes-cl->nbytes,0);
#ifdef DEBUG_LOGD
				printf("[%d] read %d bytes n=%d\n",i,lbytes,cl->nbytes);
#endif				
				if(lbytes){
					cl->nbytes+=lbytes;
#ifdef DEBUG_LOGD
					printf("[%d] AAA n[i]=%d r[i]=%d\n",i,cl->nbytes,cl->rbytes);
#endif					
					if(cl->nbytes==cl->rbytes){
#ifdef DEBUG_LOGD
						printf("[%d] struct read %d bytes\n",i,cl->nbytes);
#endif
						client_action(i,cl);
						cl->nbytes=0;
					}
				}else{
					fprintf(stderr,"%s Client #%d left\n",Logd_time_pid_info(&LOGD),i);
					FD_CLR(cl->fd,&mask);
					closesocket(cl->fd);
					UDM_FREE(cl->data.buf);
					init_client(cl,1);
				}
			}
		}
	}
}
