#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_db.h"
#include "udm_log.h"
#include "udm_spell.h"
#include "udm_xmalloc.h"
#include "udm_agent.h"
#include "udm_utils.h"
#include "udm_crossword.h"

#define RESORT_WORDS	256
#define WSIZE		1024


static int AddOneCrossWord(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc,UDM_SERVER *Server,UDM_CROSSWORD * CrossWord){
	CrossWord->pos=Doc->wordpos;

	/* Realloc memory when required  */
	if(Doc->CrossList.ncrosswords>=Doc->CrossList.mcrosswords){
		if(Doc->CrossList.mcrosswords){
			Doc->CrossList.mcrosswords+=WSIZE;
			Doc->CrossList.CrossWord=(UDM_CROSSWORD *)UdmXrealloc(Doc->CrossList.CrossWord,Doc->CrossList.mcrosswords*sizeof(UDM_CROSSWORD));
		}else{
			Doc->CrossList.mcrosswords=WSIZE;
			Doc->CrossList.CrossWord=(UDM_CROSSWORD *)UdmXmalloc(Doc->CrossList.mcrosswords*sizeof(UDM_CROSSWORD));
		}
	}

	/* Add new word */
	Doc->CrossList.CrossWord[Doc->CrossList.ncrosswords].word=strdup(CrossWord->word);
	Doc->CrossList.CrossWord[Doc->CrossList.ncrosswords].url=strdup(CrossWord->url);
	Doc->CrossList.CrossWord[Doc->CrossList.ncrosswords].weight=CrossWord->weight;
	Doc->CrossList.CrossWord[Doc->CrossList.ncrosswords].pos=CrossWord->pos;
	Doc->CrossList.ncrosswords++;
	return(0);
}

/* This function adds a normalized word form(s) into list using Ispell */
int UdmAddCrossWord(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc,UDM_SERVER *Server,UDM_CROSSWORD * CrossWord){
	char 	** forms, ** saveforms;
	int	have_digit=0;
	int	have_alpha=0;

	Doc->wordpos++;

	/* FIXME UNI: move checking to UNICODE form */
	if(Server->number_factor==0||Server->alnum_factor==0){
		char *s;
		s=CrossWord->word;
		while(*s){
			if(isdigit(*s))
				have_digit=1;
			else	
				have_alpha=1;
			if(have_digit&&have_alpha)break;
			s++;
		}
		if(have_digit){
			if(have_alpha){
				if(!Server->alnum_factor)return(0);
			}else{
				if(!Server->number_factor)return(0);
			}
		}
	}

	/* if((saveforms=forms=UdmNormalizeWord(Indexer,CrossWord->word))){ FIXME UNI*/
	if(0){
		/* Add all NORMAL forms of the word */
		while(*forms){
			/* Add only if correct words are allowed */
			if(Server->correct_factor){
				UDM_CROSSWORD cw;
				memcpy(&cw,CrossWord,sizeof(UDM_CROSSWORD));
				cw.word=*forms;
				AddOneCrossWord(Indexer,Doc,Server,&cw);
			}
			free(*forms);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMAL forms has not been found  */
		/*   then we will add the word itself  */
		/* Do it only when incorrect words are */
		/* allowed by configuration            */
		if(Server->incorrect_factor)
			AddOneCrossWord(Indexer,Doc,Server,CrossWord);
	}
	return(0);
}

void UdmFreeCrossList(UDM_CROSSLIST * CrossList) {
	size_t i;
	for(i=0;i<CrossList->ncrosswords;i++){
		free(CrossList->CrossWord[i].word);
		free(CrossList->CrossWord[i].url);
	}
	CrossList->ncrosswords=0;
	CrossList->mcrosswords=0;
	UDM_FREE(CrossList->CrossWord);
	return;
}

UDM_CROSSLIST * UdmInitCrossList(UDM_CROSSLIST * List){
	bzero(List,sizeof(*List));
	return(List);
}
