#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_doc.h"
#include "udm_utils.h"
#include "udm_doc.h"
#include "udm_word.h"
#include "udm_crossword.h"
#include "udm_hrefs.h"
#include "udm_guesser.h"
#include "udm_header.h"

UDM_DOCUMENT * UdmInitDoc(UDM_DOCUMENT * Doc){
	bzero(Doc,sizeof(UDM_DOCUMENT));
	return(Doc);
}

UDM_DOCUMENT * UdmAllocDoc(void){
	UDM_DOCUMENT * Doc=(UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT));
	bzero(Doc,sizeof(UDM_DOCUMENT));
	return(Doc);
}

static void free_doc(UDM_DOCUMENT * Result){
	UDM_FREE(Result->url);
	UDM_FREE(Result->content_type);
	UDM_FREE(Result->title);
	UDM_FREE(Result->text);
	UDM_FREE(Result->keywords);
	UDM_FREE(Result->description);
	UDM_FREE(Result->lang);
	UDM_FREE(Result->tag);
	UDM_FREE(Result->category);
	UDM_FREE(Result->headers);
	UDM_FREE(Result->subject);
	UDM_FREE(Result->from);
	UDM_FREE(Result->newsgroups);
	UDM_FREE(Result->references);
	UDM_FREE(Result->message_id);
	UDM_FREE(Result->request);
	UDM_FREE(Result->charset);
	UDM_FREE(Result->textbuf);
	UDM_FREE(Result->proxy_host);
	UdmFreeHrefList(&Result->HrefList);
	UdmFreeWordList(&Result->WordList);
	UdmFreeCrossList(&Result->CrossList);
	UdmHeaderListFree(&Result->HeaderList);
}

void UdmFreeDoc(UDM_DOCUMENT *Result){
	if(!Result)return;
	free_doc(Result);
	free(Result);
}


void UdmFreeTargets(UDM_TARGETS * T){
	size_t i;
	
	for(i=0;i<T->num_rows;i++)
		free_doc(&T->Doc[i]);
	UDM_FREE(T->Doc);
}

__INDLIB__ void UdmFreeResult(UDM_RESULT * Res){
	size_t i;
	
	if(Res){
		if(Res->Doc){
			for(i=0;i<Res->num_rows;i++){
				free_doc(&Res->Doc[i]);
			}
			UDM_FREE(Res->Doc);
		}
		UDM_FREE(Res);
	}
	return;
}

#define nonul(x)	((x)?(x):"")

int UdmDocToTextBuf(UDM_DOCUMENT * Doc){
	size_t len=10000; /* FIXME */
	
	UDM_FREE(Doc->textbuf);
	Doc->textbuf=malloc(len+1);
	
	snprintf(Doc->textbuf,len,"DI=%d\tDU=%s\tDC=%s\tDM=%li\tDT=%s\tDX=%s\tDS=%d\tDK=%s\tDD=%s\tDCRC=%d\r\n",
		Doc->url_id,
		nonul(Doc->url),
		nonul(Doc->content_type),
		Doc->last_mod_time,
		nonul(Doc->title),
		nonul(Doc->text),
		Doc->size,
		nonul(Doc->keywords),
		nonul(Doc->description),
		Doc->crc32);
	/*
	Currently unsupported 
	Doc.next_index_time=atoi
	Doc.referrer=atoi
	Doc.tag=strdup
	Doc.category=strdup
	Doc.lang=strdup
	*/
	return(0);
}

int UdmDocFromTextBuf(UDM_DOCUMENT * Doc){
	char * tok, * lt;
	if(!Doc->textbuf)return(0);

	tok=UdmGetToken(Doc->textbuf,"\t",&lt);

	while(tok){
		char * data;
		
		data=strchr(tok,'=');
		if(!data)continue;
		*data='\0';
		data++;
		
		if(!strcmp(tok,"DU")){
			UDM_FREE(Doc->url);
			Doc->url=strdup(data);
		}else
		if(!strcmp(tok,"DC")){
			UDM_FREE(Doc->content_type);
			Doc->content_type=strdup(data);
		}else
		if(!strcmp(tok,"DM")){
			Doc->last_mod_time=atoi(data);
		}else
		if(!strcmp(tok,"DT")){
			UDM_FREE(Doc->title);
			Doc->title=strdup(data);
		}else
		if(!strcmp(tok,"DX")){
			UDM_FREE(Doc->text);
			Doc->text=strdup(data);
		}else
		if(!strcmp(tok,"DS")){
			Doc->size=atoi(data);
		}else
		if(!strcmp(tok,"DK")){
			UDM_FREE(Doc->keywords);
			Doc->keywords=strdup(data);
		}else
		if(!strcmp(tok,"DD")){
			UDM_FREE(Doc->description);
			Doc->description=strdup(data);
		}else
		if(!strcmp(tok,"DS")){
			Doc->crc32=atoi(data);
		}else{
			/* Unknown field */
		}
		tok=UdmGetToken(NULL,"\t",&lt);
	}
	return 0;
}
