#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_parsehtml.h"

#include "udm_utils.h"
#include "udm_tag.h"
#include "udm_url.h"
#include "udm_filter.h"
#include "udm_log.h"
#include "udm_xmalloc.h"
#include "udm_server.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_crossword.h"
#include "udm_spell.h"
#include "udm_unicode.h"
#include "udm_sgml.h"
#include "udm_guesser.h"
#include "udm_stopwords.h"
#include "udm_crc32.h"
#include "udm_section.h"


const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t){
	const char * e , * b;
	
	t->ntoks=0;
	
	if(s == NULL && (s=*lt)==NULL)
		return NULL;

	if(!*s)return NULL;
	
	if(!strncmp(s,"<!--",4))t->type=UDM_HTML_COM;
	else	
	if(*s=='<')t->type=UDM_HTML_TAG;
	else	t->type=UDM_HTML_TXT;

	switch(t->type){
		case UDM_HTML_TAG:

			for(b=s+1;*b;){
				const char * valbeg=NULL;
				const char * valend=NULL;
				size_t nt=t->ntoks;
				
				
				/* Skip leading spaces */
				while((*b)&&strchr(" \t\r\n",*b))b++;

				if(*b=='>'){
					*lt=b+1;
					return(s);
				}

				/* Skip non-spaces, i.e. name */
				for(e=b;(*e)&&!strchr(" =>\t\r\n",*e);e++);
				
				if(t->ntoks<UDM_MAXTAGVAL)
					t->ntoks++;
				
				t->toks[nt].val=0;
				t->toks[nt].vlen=0;
				t->toks[nt].name=b;
				t->toks[nt].nlen=e-b;

				if(!strncasecmp(b,"script",6))t->script=1;
				if(!strncasecmp(b,"/script",7))t->script=0;

				if(*e=='>'){
					*lt=e+1;
					return(s);
				}

				if(!(*e)){
					*lt=e;
					return(s);
				}
				
				/* Skip spaces */
				while((*e)&&strchr(" \t\r\n",*e))e++;
				
				if(*e!='='){
					b=e;
					continue;
				}
				
				/* Skip spaces */
				for(b=e+1;(*b)&&strchr(" \r\n\t",*b);b++);
				
				if(*b=='"'){
					b++;
					
					valbeg=b;
					for(e=b;(*e)&&(*e!='"');e++);
					valend=e;
					
					b=e;
					if(*b=='"')b++;
				}else
				if(*b=='\''){
					b++;
					
					valbeg=b;
					for(e=b;(*e)&&(*e!='\'');e++);
					valend=e;
					
					b=e;
					if(*b=='\'')b++;
				}else{
					valbeg=b;
					for(e=b;(*e)&&!strchr(" >\t\r\n",*e);e++);
					valend=e;
					
					b=e;
				}
				*lt=b;
				t->toks[nt].val=valbeg;
				t->toks[nt].vlen=valend-valbeg;
			}
			break;

		case UDM_HTML_COM: /* comment */
			
			if(!UDM_STRNCASECMP(s,"<!--UdmComment-->"))
				t->comment=1;
			else
			if(!UDM_STRNCASECMP(s,"<!--/UdmComment-->"))
				t->comment=0;

			for(e=s;(*e)&&(strncmp(e,"-->",3));e++);
			if(!strncmp(e,"-->",3))*lt=e+3;
			else	*lt=e;
			break;

		case UDM_HTML_TXT: /* text */
		default:
			/* Special case when script  */
			/* body is not commented:    */
			/* <script> x="<"; </script> */
			/* We should find </script>  */
			/* in this case:             */
			
			if(t->script){
				for(e=s;*e;e++){
					if(!UDM_STRNCASECMP(e,"</script>"))
						break;
				}
			}else{
				for(e=s;(*e)&&(*e!='<');e++);
			}
			
			*lt=e;
			break;
	}
	return s;
}


/****************************************************************/

static int UdmUniParseText(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_SERVER * Server,int *ustr,int weight,char * href){
	int * lt,*tok;
	int res=IND_OK;
	UDM_CHARSET * cs;

	cs=UdmGetCharSet(Indexer->Conf->local_charset?Indexer->Conf->local_charset:"utf-8");

	tok=UdmUniGetToken(ustr,&lt);
	while(tok){
		size_t tlen;			   /* Word length          */ 
		int ures;
		int uword[UDM_MAXWORDSIZE+1];	   /* Word in UNICODE      */
		char lcsword[UDM_MAXWORDSIZE*4+1]; /* Word in LocalCharset */

		tlen=lt-tok;
		if(tlen<=UDM_MAXWORDSIZE){
			int addme=1;
			UDM_STOPWORD * stop;
			
			memcpy(uword,tok,tlen*sizeof(int));
			uword[tlen]=0;
			ures=cs->from_uni(lcsword,uword,sizeof(lcsword));
			ures=strlen(lcsword);

			/* Check MaxWordLen and MinWordLen FIXME: move to unicode form */
			if(ures>Indexer->Conf->max_word_len)addme=0;
			if(ures<Indexer->Conf->min_word_len)addme=0;

			/* Check stopword */
			if((stop=UdmIsStopWord(Indexer->Conf,lcsword))){
				addme=0;
				UdmLog(Indexer,UDM_LOG_DEBUG,"stop: '%s'-'%s'", stop->word, stop->lang);
			}

			if(addme){
				res=UdmAddWord(Indexer, Doc,Server, lcsword, weight);
				if(res!=IND_OK)return(res);
				if(href){
					UDM_CROSSWORD cw;
					cw.url=href;
					cw.weight=Server->cross_section;
					cw.pos=Doc->wordpos;
					cw.word=lcsword;
					UdmAddCrossWord(Indexer,Doc,Server,&cw);
				}
			}
		}
		tok=UdmUniGetToken(NULL,&lt);
	}
	return res;
}

/****************************************************************/

typedef struct {
	char * str;
	int weight;
	int section;
	char * href;
} UDM_TEXTITEM;

typedef struct {
	size_t nitems;
	UDM_TEXTITEM * items;
} UDM_TEXTLIST;


static void UdmTextListAdd(UDM_TEXTLIST * tlist,char * str,int weight,char * href,int section){
	if(!str)return;

	if(tlist->nitems){
		tlist->items=(UDM_TEXTITEM*)realloc(tlist->items,(tlist->nitems+1)*sizeof(UDM_TEXTITEM));
	}else{
		tlist->items=(UDM_TEXTITEM*)malloc((tlist->nitems+1)*sizeof(UDM_TEXTITEM));
	}
	
	tlist->items[tlist->nitems].str=strdup(str);
	tlist->items[tlist->nitems].weight=weight;
	tlist->items[tlist->nitems].section=section;
	tlist->items[tlist->nitems].href=href?strdup(href):NULL;

	tlist->nitems++;
	return;
}


static int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_SERVER * CurSrv,UDM_TEXTLIST * tlist, int add){
	size_t i;
	char * doccset;
	char asc[]="us-ascii";
	UDM_CHARSET * doccs;
	UDM_CHARSET * loccs;
	size_t textlen=0;
	int res=IND_OK;

	if(Indexer->Conf->LangMapList.nmaps)
		UdmGuessCharSet(Indexer,Doc);

	doccset=Doc->charset;
	if(!doccset||!*doccset)doccset=CurSrv->charset;
	if(!doccset||!*doccset)doccset=asc;
	doccs=UdmGetCharSet(doccset);
	if(!doccs)doccs=UdmGetCharSet("us-ascii");
	loccs=UdmGetCharSet(Indexer->Conf->local_charset?Indexer->Conf->local_charset:"utf-8");
	if(!loccs)loccs=UdmGetCharSet("utf-8");

	/* Now convert everything to UNICODE format and calculate CRC32 */
	Doc->crc32=0;

	for(i=0;i<tlist->nitems;i++){
		size_t len;
		int * ustr;
		int ures;
		int uspace[2]={' ',0};

		/* +2 for space and terminating 0 */
		len=strlen(tlist->items[i].str)+2;
		ustr=(int*)malloc((len)*sizeof(int));
		ures=doccs->to_uni(ustr,tlist->items[i].str,len);
		UdmUniUnescapeSgml(ustr);
		UdmUniRemoveDoubleSpaces(ustr);
		if(ures<0)ures=-ures;

		/*
		TODO for clones detection:
		Replace any separators into space to ignore 
		various pseudo-graphics, commas, semicolons
		and so on to improve clone detection quality
		*/

		if(i){
			/* Add space between porsions */
			Doc->crc32=UdmCRC32Update(Doc->crc32,(char*)(&uspace),sizeof(int));
		}
		Doc->crc32=UdmCRC32Update(Doc->crc32,(char*)ustr,ures*sizeof(int));

		switch(tlist->items[i].section){
			case UDM_SECTION_TITLE:
				UDM_FREE(Doc->title);
				len=UDM_MAXTITLESIZE+1;
				Doc->title=(char*)malloc(len);
				ures=loccs->from_uni(Doc->title,ustr,len-1);
				break;
			case UDM_SECTION_KWRD:
				UDM_FREE(Doc->keywords);
				len=UDM_MAXKEYWORDSIZE+1;
				Doc->keywords=(char*)malloc(len);
				ures=loccs->from_uni(Doc->keywords,ustr,len);
				break;
			case UDM_SECTION_DESC:
				UDM_FREE(Doc->description);
				len=UDM_MAXDESCSIZE+1;
				Doc->description=(char*)malloc(len);
				ures=loccs->from_uni(Doc->description,ustr,len);
				break;
			case UDM_SECTION_BODY:
				/* +4 to avoid attempts to fill the only one  */
				/* last byte with multibyte sequence          */
				/* as well as to add a space between portions */
				if((textlen+4)<Indexer->Conf->max_text_size){
					if(!textlen){
						UDM_FREE(Doc->text);
						Doc->text=(char*)malloc(Indexer->Conf->max_text_size+1);
					}else{
						/* Add space */
						ures=loccs->from_uni(Doc->text+textlen,uspace,2);
						textlen=strlen(Doc->text);
					}
					len=Indexer->Conf->max_text_size-textlen;
					ures=loccs->from_uni(Doc->text+textlen,ustr,len);
					textlen=strlen(Doc->text);
				}
				break;
		}
		UdmUniStrToLower(ustr);
		if(add){
			res=UdmUniParseText(Indexer,Doc,CurSrv,ustr,tlist->items[i].weight,tlist->items[i].href);
			if(res!=IND_OK)return(res);
		}
		UDM_FREE(tlist->items[i].str);
		UDM_FREE(tlist->items[i].href);
		free(ustr);
	}
	return res;
}


/**************************** Built-in Parsers ***************************/

static int UdmParseURLText(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_SERVER * CurSrv, UDM_URL * CurURL, UDM_TEXTLIST * tlist){
	char str[UDM_URLSIZE]="";

	if(CurSrv->proto_section) {
		strcpy(str,CurURL->schema);
		UdmTextListAdd(tlist,str,CurSrv->proto_section,NULL,UDM_SECTION_PROTO);
	}
	if(CurSrv->host_section) {
		strcpy(str,CurURL->hostname);
		UdmTextListAdd(tlist,str,CurSrv->host_section,NULL,UDM_SECTION_HOST);
	}
	if(CurSrv->path_section) {
		UdmUnescapeCGIQuery(str,CurURL->path);
		UdmTextListAdd(tlist,str,CurSrv->path_section,NULL,UDM_SECTION_PATH);
	}
	if(CurSrv->file_section) {
		UdmUnescapeCGIQuery(str,CurURL->filename);
		UdmTextListAdd(tlist,str,CurSrv->file_section,NULL,UDM_SECTION_FILE);
	}
	
	{
		size_t i;
		for(i=0;i<Doc->HeaderList.nheaders;i++){
			UDM_SECTION * Sec;

			if((Sec=UdmSectionListFind(&Indexer->Conf->SectionList,Doc->HeaderList.Headers[i].name))){
				char * c;
				c=strdup(Doc->HeaderList.Headers[i].value);
				UdmTextListAdd(tlist,c,Sec->section,NULL,Sec->section);
				free(c);
			}
		}
	}
	
	return 0;
}


int UdmParseText(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,UDM_SERVER * CurSrv, UDM_URL * CurURL){
	char *s, *lt;
	UDM_TEXTLIST tlist;
	int res=IND_OK;

	bzero(&tlist,sizeof(tlist));

	if(CurSrv->body_section&&Doc->content){
		s=UdmGetToken(Doc->content, "\r\n", &lt);
		while(s){
			UdmTextListAdd(&tlist,s,CurSrv->body_section,NULL,UDM_SECTION_BODY);
			if(Indexer->Conf->LangMapList.nmaps)
				UdmBuildLangMap(&Doc->LangMap,s,strlen(s));
			s=UdmGetToken(NULL, "\r\n", &lt);
		}
	}
	if(Doc->subject){ /* FIXME: add SubjWeight*/ 
		UdmTextListAdd(&tlist,Doc->subject,CurSrv->title_section,NULL,UDM_SECTION_TITLE);
	}
	UdmParseURLText(Indexer,Doc,CurSrv,CurURL, &tlist);
	res=UdmPrepareWords(Indexer,Doc,CurSrv,&tlist,1);
	UDM_FREE(tlist.items);
	return(res);
}

#define ISTAG(n,s)	(!strncasecmp(tag->toks[n].name,s,tag->toks[n].nlen)&&(tag->toks[n].nlen==strlen(s)))

static char * strndup(const char * str,size_t len){
	char * res;
	res=(char*)malloc(len+1);
	strncpy(res,str,len);
	res[len]='\0';
	return res;
}


static int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc,UDM_SERVER * CurSrv,UDM_TEXTLIST * tlist,UDM_SECTIONLIST * SecList,UDM_URL *CurURL){
	int opening;
	char name[128];
	char * n;
	char *metaname=NULL;
	char *metacont=NULL;
	char *href=NULL;
	size_t i;

	if(!tag->ntoks)return(0);
	if(!tag->toks[0].name)return(0);
	if(tag->toks[0].nlen>sizeof(name)-1)return(0);

	strncpy(name,tag->toks[0].name,tag->toks[0].nlen);
	name[tag->toks[0].nlen]='\0';
	
	for(i=0;i<tag->ntoks;i++){
		if(ISTAG(i,"name")){
			metaname=strndup(tag->toks[i].val,tag->toks[i].vlen);
		}else
		if(ISTAG(i,"http-equiv")){
			metaname=strndup(tag->toks[i].val,tag->toks[i].vlen);
		}else
		if(ISTAG(i,"content")){
			metacont=strndup(tag->toks[i].val,tag->toks[i].vlen);
		}else
		if(ISTAG(i,"href")){
			/* A, LINK, AREA*/
			href=strndup(tag->toks[i].val,tag->toks[i].vlen);
		}else
		if(ISTAG(i,"src")){
			/* IMG, FRAME */
			href=strndup(tag->toks[i].val,tag->toks[i].vlen);
		}
	}
	
	for(n=name;*n;*n=tolower(*n),n++);
	
	if(name[0]=='/'){
		opening=0;
		memmove(name,name+1,strlen(name+1)+1);
	}else{
		opening=1;
	}

	/* Let's find tag name in order of frequency */

	if(!strcmp(name,"a")){
		UDM_FREE(tag->lasthref);			/*117941*/
	}else
	if(!strcmp(name,"title"))	tag->title=opening;	/*6192*/
	else
	if(!strcmp(name,"body"))	tag->body=opening;	/*5146*/
	else
	if((!strcmp(name,"meta"))&&(metaname)&&(metacont)){ 
		UDM_SECTION * Sec;
		if((Sec=UdmSectionListFind(SecList,metaname))){
			if(!tag->comment){
				char * c;
				c=strdup(metacont);
				UdmTextListAdd(tlist,c,Sec->section,NULL,Sec->section);
				free(c);
			}
		}
		if((!strcasecmp(metaname,"Content-Type"))&&(!Doc->charset||!*Doc->charset)){
			char *p;
			if((p=strstr(metacont,"charset="))){
				const char *cs = UdmCharsetCanonicalName(p+8);
				UDM_FREE(Doc->charset);
				Doc->charset=strdup(cs?cs:(p+8));
			}
		}else
		if(!strcasecmp(metaname,"refresh")){
			/* Format: "10; Url=http://something/" */
			/* URL can be written in different     */
			/* forms: URL, url, Url and so on      */
			
			if((href=strchr(metacont,'='))){
				if((href>=metacont+3)&&(!UDM_STRNCASECMP(href-3,"URL="))){
					href=strdup(href+1);
				}else{
					UDM_FREE(href);
				}
			}
 		}else
		if(!strcasecmp(metaname,"keywords")){
			UdmUnescapeSgmlStr(metacont);
			if(!tag->comment){
				UdmTextListAdd(tlist,metacont,CurSrv->kwrd_section,NULL,UDM_SECTION_KWRD);
			}
		}else
		if(!strcasecmp(metaname,"description")){
			UdmUnescapeSgmlStr(metacont);
			if(!tag->comment){
				UdmTextListAdd(tlist,metacont,CurSrv->desc_section,NULL,UDM_SECTION_DESC);
			}
		}else
		if(!strcasecmp(metaname,"robots")&&(CurSrv->use_robots)&&(metacont)){
			char * lt;
			char * rtok;
					
			rtok=UdmGetToken(metacont," \r\n\t",&lt);
			while(rtok){
				if(!strcasecmp(rtok,"ALL")){
					/* Set Server parameters */
					tag->follow=CurSrv->follow;
					tag->index=CurSrv->index;
				}else
				if(!strcasecmp(rtok,"NONE")){
					tag->follow=UDM_FOLLOW_NO;
					tag->index=0;
				}else
				if(!strcasecmp(rtok,"NOINDEX"))
					tag->index=0;
				else
				if(!strcasecmp(rtok,"NOFOLLOW"))
					tag->follow=UDM_FOLLOW_NO;
				else
				if(!strcasecmp(rtok,"INDEX"))
					tag->index=1;
				else
				if(!strcasecmp(rtok,"FOLLOW"))
					tag->follow=CurSrv->follow;
				rtok=UdmGetToken(NULL," \r\n\t",&lt);
			}
		}
	}
	else	if(!strcmp(name,"script"))	tag->script=opening;
	else	if(!strcmp(name,"style"))	tag->style=opening;
	else	if(!strcmp(name,"noindex"))	tag->comment=opening;
	else	
	if((!strcmp(name,"base"))&&(href)){
		/* <BASE HREF="xxx"> stuff            */
		/* Check that URL is properly formed  */
		/* baseURL is just temporary variable */
		/* If parsing  fails we'll use old    */
		/* base href, passed via CurURL       */
		/* Note that we will not check BASE     */
		/* if delete_no_server is unset         */
		/* This is  actually dirty hack. We     */
		/* must check that hostname is the same */

		if((CurSrv->delete_no_server)||(CurSrv->follow==UDM_FOLLOW_WORLD)){
			UDM_URL baseURL;
			int parse_res;
					
			if(!(parse_res=UdmParseURL(&baseURL,href))){
				UdmParseURL(CurURL,href);
			}else{
				/*
				switch(parse_res){
				case UDM_URL_LONG:
					UdmLog(Indexer,UDM_LOG_ERROR,"BASE HREF too long: '%s'",tag.href);
					break;
				case UDM_URL_BAD:
				default:
					UdmLog(Indexer,UDM_LOG_ERROR,"Error in BASE HREF URL: '%s'",tag.href);
				}
				*/
			}
		}
	}

	if((href)&&(tag->follow!=UDM_FOLLOW_NO)){
		UDM_URL newURL;
		int parse_res;
			
		if(!(parse_res=UdmParseURL(&newURL,href))){
			char * newschema;
			UDM_DOCUMENT Href;
			char str[UDM_URLSIZE];
			char pathfile[UDM_URLSIZE];
					
			if(newURL.schema[0])newschema=newURL.schema;
			else	newschema=CurURL->schema;

			sprintf(pathfile,"%s%s",newURL.path[0]?newURL.path:CurURL->path,newURL.filename);
			UdmURLNormalizePath(pathfile);

			if(!strcasecmp(newschema,"file")||!strcasecmp(newschema,"htdb")){
				sprintf(str,"%s:%s",newschema,pathfile);
			}else{
				sprintf(str,"%s://%s%s",newschema,newURL.hostinfo[0]?newURL.hostinfo:CurURL->hostinfo,pathfile);
			}

			if(!UDM_STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
				*(strstr(str,";type"))=0;

			Href.referrer=Doc->url_id;
			Href.stored=0;
			Href.tag=NULL;
			Href.category=NULL;
			Href.hops=Doc->hops+1;
			Href.url=str;
			UdmAddHref(&Doc->HrefList,&Href);

			/* For crosswords */
			UDM_FREE(tag->lasthref);
			tag->lasthref=strdup(str);
		}else{
			/*
			switch(parse_res){
				case UDM_URL_LONG:
					UdmLog(Indexer,UDM_LOG_DEBUG,"URL too long: '%s'",href);
					break;
				case UDM_URL_BAD:
				default:
					UdmLog(Indexer,UDM_LOG_DEBUG,"Error in URL: '%s'",href);
			}
			*/
		}
	}
	UDM_FREE(metaname);
	UDM_FREE(metacont);
	UDM_FREE(href);

	return 0;
}


#define MAXSTACK	1024

typedef struct {
	size_t len;
	char * ofs;
} UDM_TAGSTACK;

int UdmParseHtml(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_SERVER *CurSrv,UDM_URL *CurURL){
	int hrefonly=0;
	const char *htok, *last;
	char reason[1024]="";
	char * filter=NULL;
	UDM_TEXTLIST tlist;
	int res=IND_OK;
	UDM_HTMLTOK tag;
	char   stack[MAXSTACK*8]="";
	char   * sbot=stack;
	size_t nstack=0;
	UDM_TAGSTACK Stack[MAXSTACK];

	bzero(&tlist,sizeof(tlist));
	bzero(&tag,sizeof(tag));

	tag.follow=CurSrv->follow;
	tag.index=CurSrv->index;

	filter=strdup(Doc->url);
	hrefonly=(UDM_HREFONLY==UdmFindFilter(Indexer->Conf,filter,reason));
	free(filter);

	htok=Doc->content;
	
	htok=UdmHTMLToken(Doc->content,&last,&tag);
	
	while(htok){
		char * tmp=NULL;
		size_t tmplen;

		switch(tag.type){
			
			case UDM_HTML_COM:

				
				break;

			case UDM_HTML_TXT:

				tmplen=(last-htok)+6;
				tmp=(char*)malloc((size_t)(last-htok+1));
				strncpy(tmp,htok,(size_t)(last-htok));
				tmp[last-htok]='\0';
				
				if(Indexer->Conf->LangMapList.nmaps)
					UdmBuildLangMap(&Doc->LangMap,tmp,strlen(tmp));

				if(!tag.comment&&tag.body&&!tag.script&&!tag.style){
					UdmTextListAdd(&tlist,tmp,CurSrv->body_section,tag.lasthref,UDM_SECTION_BODY);
				}
				if(!tag.comment&&tag.title){
					UdmTextListAdd(&tlist,tmp,CurSrv->title_section,tag.lasthref,UDM_SECTION_TITLE);
				}
				free(tmp);
				break;
		
			case UDM_HTML_TAG:
				
				if(0){
					size_t i;
					char str[100];
					strncpy(str,tag.toks[0].name,tag.toks[0].nlen);
					str[tag.toks[0].nlen]='\0';
					fprintf(stderr,"T=%s\n",str);
					
					switch(tag.toks[0].name[0]){
						case '!':
							
							break;
						case '/':
							for(i=nstack;i>0;i--){
								if(Stack[i-1].len+1==tag.toks[0].nlen){
									if(!strncasecmp(Stack[i-1].ofs,tag.toks[0].name+1,Stack[i-1].len)){
										nstack=i-1;
										sbot=Stack[i-1].ofs;
										if(sbot>stack)sbot--;
										*sbot='\0';
										break;
									}
								}
							}
							break;
						default:
							if(*stack){
								strcat(sbot,".");
								sbot++;
							}
							strncpy(sbot,tag.toks[0].name,tag.toks[0].nlen);
							sbot[tag.toks[0].nlen]='\0';
							Stack[nstack].ofs=sbot;
							Stack[nstack].len=tag.toks[0].nlen;
							nstack++;
							sbot=sbot+strlen(sbot);
						
					}
					fprintf(stderr,"S=%s\n\n",stack);
				}
				UdmHTMLParseTag(&tag,Doc,CurSrv,&tlist,&Indexer->Conf->SectionList,CurURL);
				break;
		}
		htok=UdmHTMLToken(NULL,&last,&tag);
		
	}
	UdmParseURLText(Indexer,Doc,CurSrv,CurURL, &tlist);
	res=UdmPrepareWords(Indexer,Doc,CurSrv,&tlist,(!hrefonly&&tag.index));
	UDM_FREE(tlist.items);
	UDM_FREE(tag.lasthref);	
	return(res);
}
