#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include <udm_config.h>
#include <udmsearch.h>

/*
	TODO:
		- categories
		- pass tag,cat,ul arguments
*/

#ifndef UDM_CONF_DIR
#define UDM_CONF_DIR "/usr/local/mnogosearch/etc/"
#endif

/**************************************************/

#define T_UNKNOWN	 0
#define T_TOP		 1
#define T_BOTTOM	 2
#define T_RESTOP	 3
#define T_RESBOT	 4
#define T_RESULT	 5
#define T_ERROR		 6
#define T_NOTFOUND	 7
#define T_CLONE		 8
#define T_NAVLEFT	 9
#define T_NAVBAR0	10
#define T_NAVBAR1	11
#define T_NAVRIGHT	12
#define T_INCLUDE	13
#define T_NAVLEFT_NOP	14
#define T_NAVRIGHT_NOP	15
#define T_NOQUERY	16
#define T_VARIABLES	17
#define T_MAX		18

#define MAX_FORMAT	10

typedef struct {
	int type;
	const char * begin;
	const char * end;
} TPART;

const TPART part[]= 
{
	{T_TOP,		"<!--top-->",		"<!--/top-->"},
	{T_BOTTOM,	"<!--bottom-->",	"<!--/bottom-->"},
	{T_RESTOP,	"<!--restop-->",	"<!--/restop-->"},
	{T_RESBOT,	"<!--resbot-->",	"<!--/resbot-->"},
	{T_RESULT,	"<!--res-->",		"<!--/res-->"},
	{T_NOTFOUND,	"<!--notfound-->",	"<!--/notfound-->"},
	{T_ERROR,	"<!--error-->",		"<!--/error-->"},
	{T_CLONE,	"<!--clone-->",		"<!--/clone-->"},
	{T_NAVLEFT,	"<!--navleft-->",	"<!--/navleft-->"},
	{T_NAVBAR1,	"<!--navbar1-->",	"<!--/navbar1-->"},
	{T_NAVBAR0,	"<!--navbar0-->",	"<!--/navbar0-->"},
	{T_NAVRIGHT,	"<!--navright-->",	"<!--/navright-->"},
	{T_NAVLEFT_NOP,	"<!--navleft_nop-->",	"<!--/navleft_nop-->"},
	{T_NAVRIGHT_NOP,"<!--navright_nop-->",	"<!--/navright_nop-->"},
	{T_NOQUERY,	"<!--noquery-->",	"<!--/-->"},
	{T_VARIABLES,	"<!--variables",	"/variables-->"},
	{0,}
};

typedef struct {
	int nformats;
	char * template[MAX_FORMAT];
} TMPL;



/******************* Template functions ********************/

static void InitTmpl(TMPL * t){
	int i,j;
	
	for(i=0;i<T_MAX;i++){
		t[i].nformats=0;
		for(j=0;j<MAX_FORMAT;j++)
			t[i].template[j]=NULL;
	}
}

static size_t out_string(FILE * stream,char * dst,const char * src){
	if(src){
		if(stream)fputs(src,stream);
		if(dst){
			strcat(dst,src);
			return(strlen(src));
		}
	}
	return(0);
}

static char * HiLightDup(const char * src,const char * beg, const char * end){
	size_t len=1;
	size_t blen=strlen(beg);
	size_t elen=strlen(end);
	const char * s;
	char  * res, *d;
	
	for(s=src;*s;s++){
		switch(*s){
			case '\2':
				len+=blen;
				break;
			case '\3':
				len+=elen;
				break;
			default:
				len++;
		}
	}
	res=(char*)malloc(len);
	for(s=src,d=res;*s;s++){
		switch(*s){
			case '\2':
				strcpy(d,beg);
				d+=blen;
				break;
			case '\3':
				strcpy(d,end);
				d+=elen;
				break;
			default:
				*d=*s;
				d++;
		}
	}
	*d='\0';
	return(res);
}

static size_t PrintTextTemplate(FILE * stream,char * dst,UDM_VARLIST * vars,const char * templ){
	const char * s;
	size_t dlen=0;
	const char * HlBeg=UdmFindStrVar(vars,"HlBeg","");
	const char * HlEnd=UdmFindStrVar(vars,"HlEnd","");

	for(s=templ;*s;s++){
		int type=0;
		char *value=NULL, *eval=NULL, *eval1=NULL;
		char empty[]="";

		if(*s=='$'){
			const char * vbeg=NULL, * vend;
			
			if(!strncmp(s,"$(",2)){vbeg=s+2;type='(';}
			else	if(!strncmp(s,"$%(",3)){vbeg=s+3;type='%';}
			else	if(!strncmp(s,"$&(",3)){vbeg=s+3;type='&';}

			if((type)&&(vend=strchr(s,')'))){
				UDM_VAR * var;
				size_t len;
				char name[100]="";
				
				len=(vend-vbeg);
				if(len>=sizeof(name))len=sizeof(name);
				strncpy(name,vbeg,len);name[len]='\0';
				if((var=UdmFindVar(vars,name))){
					value=var->val;
					if(!value)value=empty;
				}else{	
					value=empty;
				}
				
				s=vend;
			}else{
				type=0;
			}
		}
		switch(type){
			case '(': 
				dlen+=out_string(stream,dst+dlen,value);
				break;
			case '&':
				eval=UdmHtmlSpecialChars(value);
				eval1=HiLightDup(eval,HlBeg,HlEnd);
				dlen+=out_string(stream,dst+dlen,eval1);
				free(eval);
				free(eval1);
				break;
			case '%':
				eval=(char*)malloc(strlen(value)*3);
				UdmEscapeURL(eval,value);
				dlen+=out_string(stream,dst+dlen,eval);
				free(eval);break;
			default:	/* One character */
				if((stream)&&(*s))fputc(*s,stream);
				if(dst){
					dst[dlen]=*s;
					dlen++;
					dst[dlen]='\0';
				}
		}
		/*if(dst)fprintf(stderr,"dst='%s'\n",dst);*/
	}
	return dlen;
}

#define T_OPT	1
#define T_INP	2

static char * GetHtmlTok(const char * src,const char ** lt){
	char * res;
	size_t len;
	if((!src)&&!(src=*lt))return(NULL);
	if(*src=='<'){
		/* Find ">" and skip "<" */
		if((*lt=strchr(src,'>')))(*lt)++;
	}else{
		/* Find tag beginning */
		*lt=strchr(src,'<');
	}
	if(!(*lt)){
		/* Last token */
		res=strdup(src);
	}else{
		/* Token in the middle */
		len=(*lt)-src;
		res=(char*)malloc(len+2);
		strncpy(res,src,len);
		res[len]='\0';
	}
	return(res);
}

static void PrintHtmlTemplate(UDM_AGENT * Agent,FILE * stream,char * dst,UDM_VARLIST * vars, const char * template){
	const char *lt;
	char  *tok;
	size_t dlen=0;
	
	tok=GetHtmlTok(template,&lt);
	while(tok){
		int type=0;
		if(!(UDM_STRNCASECMP(tok,"<OPTION"))){
			type=T_OPT;	
		}else
		if(!(UDM_STRNCASECMP(tok,"<INPUT"))){
			type=T_INP;
		}else
		if(!UDM_STRNCASECMP(tok,"<!INCLUDE")){
			if(Agent){
				int size;
				UDM_TAG tag;
				UDM_URL url;
				UDM_DOCUMENT * Inc=UdmAllocDoc();

				if(!Agent->buf)Agent->buf=malloc(Agent->Conf->max_doc_size);

				UdmParseTag(&tag,tok);
				if(tag.content){
					UdmParseURL(&url,tag.content);
					Inc->request=malloc(strlen(tag.content)+512);
					sprintf(Inc->request,"GET %s%s HTTP/1.0\r\nHost: %s\r\n\r\n",url.path,url.filename,url.hostname);
					size=UdmGetURL(Agent,Inc,&url);
					if(size>0)Agent->buf[size]='\0';
					UdmParseHTTPResponse(Agent,Inc);
					if(Inc->content){
						if(stream){
							fprintf(stream,"%s",Inc->content);
						}else{
							/* FIXME */
						}
					}
				}
			}
		}else{
			dlen+=PrintTextTemplate(stream,dst+dlen,vars,tok);
		}
		if(type){
			char * opt;
			UDM_TAG  tag;
			UDM_VAR * var=NULL;
			char * vname=NULL;
			
			opt=(char*)malloc(strlen(tok)+200);
			strcpy(opt,tok);
			UdmParseTag(&tag,opt);
			if((vname=tag.selected)){
				vname=UdmTrim(vname,"$()");
				var=UdmFindVarWithValue(vars,vname,tag.value?tag.value:"");
			}
			if(type==T_OPT){
				sprintf(opt,"<OPTION VALUE=\"%s\"%s>",tag.value?tag.value:"",var?" SELECTED":"");
			}else{
				sprintf(opt,"<INPUT TYPE=\"%s\" NAME=\"%s\" VALUE=\"%s\"%s>",tag.type?tag.type:"",tag.name?tag.name:"",tag.value?tag.value:"",var?" CHECKED":"");
			}
			dlen+=PrintTextTemplate(stream,dst+dlen,vars,opt);
			UdmFreeTag(&tag);
			free(opt);
		}
		
		
		free(tok);
		tok=GetHtmlTok(NULL,&lt);
	}
}

static void PrintTemplate(UDM_AGENT * Agent,FILE * stream,char * dst,UDM_VARLIST * vars, TMPL * t, int where){ 	
	if(dst)*dst='\0';
	if(t[where].nformats){
		char * tmpl=NULL;
		int format;
		format=UdmFindIntVar(vars,"o",0);
		if(format>=t[where].nformats)format=t[where].nformats-1;
		if(format<0)format=0;
		tmpl=t[where].template[format];
		PrintHtmlTemplate(Agent,stream,dst,vars,tmpl);
	}
}


/* Load template  */
static int LoadTemplate(UDM_ENV * Env,UDM_VARLIST * vars,const char * tname, TMPL * tmpl){
	FILE * file;
	char str[1024];
	int where=0;

	if(!(file=fopen(tname,"r"))){
		snprintf(Env->errstr,sizeof(Env->errstr)-1,"Unable to open template '%s': %s",tname,strerror(errno));
		return(1);
	}

	while(fgets(str,sizeof(str),file)){
		int i;
		int done=0;
		
		for(i=0;part[i].type;i++){
			if(!UDM_STRNCASECMP(str,part[i].begin)){
				where=part[i].type;
				done=1;
				break;
			}
			if(!UDM_STRNCASECMP(str,part[i].end)){
				tmpl[where].nformats++;
				where=0;
				done=1;
				break;
			}
		}
		if(done)continue;
		if(where==T_VARIABLES){
			char *tok,*s,*lt;
			if(!UDM_STRNCMP(str,"-->"))continue;
			s=UdmTrim(str," \r\n");
			if(!*s)continue;
			if(*s=='#')continue;

			if((tok=UdmGetToken(str," \t\r\n",&lt))){
				char * arg=NULL;
					
				if(!strcasecmp(str,"Affix")){
					char aname[1024];
					char * args[5];
					size_t narg=0;
					
					while((tok)&&(narg<5)){
						args[narg++]=tok;
						tok=UdmGetToken(NULL," \t",&lt);
					}
					if(narg!=4){
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Affix command");
						return(1);
					}
					if(args[3][0]=='/')strncpy(aname,args[3],sizeof(aname)-1);
					else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,args[3]);
					if(UdmImportAffixes(Env,args[1],args[2],aname)){
						Env->errcode=1;
						snprintf(Env->errstr,sizeof(Env->errstr)-1,"Can't load affix :%s",aname);
						return(1);
					}
				}else
				if(!strcasecmp(str,"Spell")){
					char aname[1024];
					char * args[5];
					size_t narg=0;

					while((tok)&&(narg<5)){
						args[narg++]=tok;
						tok=UdmGetToken(NULL," \t",&lt);
					}
					if(narg!=4){
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Spell command");
						return(1);
					}
					if(args[3][0]=='/')strncpy(aname,args[3],sizeof(aname)-1);
					else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,args[3]);
					if(UdmImportDictionary(Env,args[1],args[2],aname,0,"")){
						Env->errcode=1;
						snprintf(Env->errstr,sizeof(Env->errstr)-1,"Can't load dictionary :%s",aname);
						return(1);
					}
				}else
				if(!strcasecmp(str,"Synonym")){
					char aname[1024];
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg){
						if(arg[0]=='/')strncpy(aname,arg,sizeof(aname)-1);
						else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
						if(UdmSynonymListLoad(Env,aname)){
							Env->errcode=1;
							return(1);
						}
					}else{
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Synonym command");
						return(1);
					}
				}else
				if(!strcasecmp(str,"SearchdAddr")){
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg)UdmSDCLientListAdd(&Env->sdcl,arg);
				}else
				if((str[0]=='R'||str[0]=='r')&&(str[1]>='0')&&(str[1]<='9')){
					float r;
					int ir;
					arg=UdmGetToken(NULL," =\t\r\n",&lt); 
					if(arg){
						r=atoi(arg);
						srand((unsigned)time(0));
						r=r*rand()/RAND_MAX; ir=r;
						UdmAddIntVar(vars,str,ir,UDM_VARSRC_GLOBAL);
					}
				}else
				if(!strcasecmp(str,"HlBeg")){
					UdmAddStrVar(vars,"HlBeg",lt,UDM_VARSRC_GLOBAL);
				}else
				if(!strcasecmp(str,"HlEnd")){
					UdmAddStrVar(vars,"HlEnd",lt,UDM_VARSRC_GLOBAL);
				}else
				if(!strcasecmp(str,"Limit")){
					char * sc, * nm;
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if((sc=strchr(arg,':'))){
						*sc='\0';sc++;
						nm=(char*)malloc(strlen(arg)+1+strlen("Limit-"));
						sprintf(nm,"Limit-%s",arg);
						UdmAddStrVar(vars,nm,sc,UDM_VARSRC_GLOBAL);
						free(nm);
					}
				}else
				if(!strcasecmp(str,"Alias")&&(arg)){
					char * arg1;
					
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					arg1=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg1){
						UDM_ALIAS Alias;
						Alias.match.pattern=arg;
						Alias.replace=arg1;
						Alias.match.match_type=0;
						Alias.match.case_sense=0;
						UdmAliasListAdd(&Env->AliasList,&Alias,Env->errstr,sizeof(Env->errstr));
					}
				}else{
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					UdmReplaceStrVar(vars,tok,arg,UDM_VARSRC_GLOBAL);
				}
			}
		}else{
			char * t;
			t=tmpl[where].template[tmpl[where].nformats];
			if(!t){
				t=strdup(str);
			}else{
				t=(char*)realloc(t,strlen(t)+strlen(str)+1);
				strcat(t,str);
			}
			tmpl[where].template[tmpl[where].nformats]=t;
		}
	}
	fclose(file);
	if(Env->nspell) {
		UdmSortDictionary(Env);
		UdmSortAffixes(Env);
	}
	UdmSynonymListSort(&Env->SynList);
	return 0;
}

static void FreeTemplate(TMPL * tmpl){
	int t;
	for(t=0;t<T_MAX;t++){
		int f;
		for(f=0;f<tmpl[t].nformats;f++){
			UDM_FREE(tmpl[t].template[f]);
		}
	}
}

/******************** Misc functions *********************************/

static char * BuildPageURL(UDM_VARLIST * vars,char * dst){
	int i,nargs=0;
	char * end=dst;

	strcpy(dst,UdmFindStrVar(vars,"self",""));
	end=dst+strlen(dst);

	for(i=0;i<vars->nvars;i++){
		if(vars->var[i].scope==UDM_VARSRC_QSTRING){
			strcpy(end,nargs?"&":"?");
			end++;
			UdmEscapeURL(end,vars->var[i].name);
			end=end+strlen(end);
			strcpy(end,"=");end++;
			UdmEscapeURL(end,vars->var[i].val);
			end=end+strlen(end);
			nargs++;
		}
	}
	return NULL;
}


/*****************************************************************/

int main(int argc, char ** argv){
	const char * env, * charset;
	char template_name[1024]="";
	char query_string[1024]="";
	const char * qw=NULL;
	char self[1024]="";
	TMPL tmpl[T_MAX];
	int res;
	int httpd=0;
	UDM_ENV    * Env;
	UDM_AGENT  * Agent;
	UDM_RESULT * Res;
	UDM_CHARSET * cs=NULL;
	
	/* Output Content-type if under HTTPD	 */
	/* Some servers do not pass QUERY_STRING */
	/* if the query was empty, so check	 */
	/* REQUEST_METHOD too     to be safe     */
	
	if(getenv("QUERY_STRING")||getenv("REQUEST_METHOD")){
		httpd=1;
	}
	
	UdmInit();
	InitTmpl(tmpl);
	Env=UdmAllocEnv();
	Env->vars=UdmAllocVarList();

	/* Detect self and template name */
	if((env=getenv("UDMSEARCH_TEMPLATE")))
		strncpy(template_name,env,sizeof(template_name)-1);

	if((env=getenv("QUERY_STRING"))){
		strncpy(query_string,env,sizeof(query_string)-1);
		if((env=getenv("REDIRECT_STATUS"))){

			/* Check Apache internal redirect  */
			/* via   "AddHandler" and "Action" */

			strncpy(self,(env=getenv("REDIRECT_URL"))?env:"search.cgi",sizeof(self)-1);
			if(!template_name[0]){
				strncpy(template_name,(env=getenv("PATH_TRANSLATED"))?env:"",sizeof(template_name)-1);
			}
		}else{
			/* CGI executed without Apache internal redirect */

			/* Detect $Self variable with OS independant SLASHES */
			strncpy(self,(env=getenv("SCRIPT_NAME"))?env:"search.cgi",sizeof(self)-1);

			if(!template_name[0]){
				char *s,*e;
				
				/*This is with OS specific SLASHES */
				env=((env=getenv("SCRIPT_FILENAME"))?env:"search.cgi");

				if(strcmp(UDM_CONF_DIR,".")){
					/* Take from the config directory */
					snprintf(template_name,sizeof(template_name)-1,"%s/%s", UDM_CONF_DIR,(s=strrchr(env,UDMSLASH))?(s+1):(self));
				}else{
					/* Take from the current directory */
					strncpy(template_name,env,sizeof(template_name)-1);
				}

				/* Find right slash if it presents */
				s=((s=strrchr(template_name,UDMSLASH))?s:template_name);

				/* Find .cgi substring */
				if((e=strstr(s,".cgi"))){
					/* Replace ".cgi" with ".htm" */
					e[1]='h';e[2]='t';e[3]='m';
				}else{
					strcpy(template_name,"search.htm");
				}
			}
		}
	}else{
		/* Executed from command line     */
		/* or under server which does not */
		/* pass an empty QUERY_STRING var */
		if(argv[1])snprintf(query_string,sizeof(query_string)-1,"q=%s",argv[1]);
		if(!template_name[0])snprintf(template_name,sizeof(template_name),"%s/%s", UDM_CONF_DIR,"search.htm");
	}

	bzero(tmpl,sizeof(tmpl));
	res=LoadTemplate(Env,Env->vars,template_name,tmpl);
	if(res){
		if(httpd)printf("Content-Type: text/plain\r\n\r\n");
		printf("%s\n",Env->errstr);
		FreeTemplate(tmpl);
		UdmFreeEnv(Env);
		return(0);
	}
	
	Env->DBAddr=strdup(UdmFindStrVar(Env->vars,"DBAddr","mysql://foo:bar@localhost/mnogosearch/"));
	UdmEnvSetDBMode(Env,UdmFindStrVar(Env->vars,"DBMode","single"));

	Agent=UdmAllocAgent(Env,0,UDM_OPEN_MODE_READ);
	{
		size_t i;
		for(i=0;i<Env->sdcl.nclients;i++){
			UdmSDCLientListAdd(&Agent->sdcl,Env->sdcl.Clients[i].addr);
		}
	}
	
	UdmAddStrVar(Env->vars,"QUERY_STRING",query_string,UDM_VARSRC_GLOBAL);
	UdmAddStrVar(Env->vars,"self",self,UDM_VARSRC_GLOBAL);
	
	UdmParseQueryString(Agent,Env->vars,query_string);
	
	
	charset=UdmFindStrVar(Env->vars,"BrowserCharset","utf-8");
	cs=UdmGetCharSet(charset);

	if(httpd){
		if(!cs){
			printf("Content-Type: text/plain\r\n\r\n");
			printf("Unknown BrowserCharset '%s' in template '%s'\n",charset,template_name);
			exit(0);
		}else{
			printf("Content-type: text/html; charset=%s\r\n\r\n",charset);
		}
	}else{
		if(!cs){
			printf("Unknown BrowserCharset '%s' in template '%s'\n",charset,template_name);
			exit(0);
		}
	}

	/* These parameters taken fro "variable section of template"*/
	if((env=UdmFindStrVar(Env->vars,"VarDir",NULL))){
		snprintf(Agent->Conf->vardir,sizeof(Agent->Conf->vardir)-1,"%s%s",env,UDMSLASHSTR);
	}
	

	Agent->page_size=UdmFindIntVar(Env->vars,"ps",20);
	Agent->page_number=UdmFindIntVar(Env->vars,"np",0);
	UdmAddIntVar(Env->vars,"pn",(int)Agent->page_number*Agent->page_size,UDM_VARSRC_GLOBAL);



	/* Now start displaying template*/
	PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_TOP);

	qw=UdmFindStrVar(Env->vars,"q","");
	if(!qw[0]){
		PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_NOQUERY);
	}else{
		if((Res=UdmFind(Agent,qw))){
			char search_time[100];
			UdmAddIntVar(Env->vars,"first",(int)Res->first,UDM_VARSRC_GLOBAL);
			UdmAddIntVar(Env->vars,"last",(int)Res->last,UDM_VARSRC_GLOBAL);
			UdmAddIntVar(Env->vars,"total",(int)Res->total_found,UDM_VARSRC_GLOBAL);
			UdmAddStrVar(Env->vars,"W",Res->wordinfo,UDM_VARSRC_GLOBAL);
			sprintf(search_time,"%.3f",((double)Res->work_time)/1000);
			UdmAddStrVar(Env->vars,"SearchTime",search_time,UDM_VARSRC_GLOBAL);
			
			if(Res->num_rows){
				int i,page1,page2,npages;
				int ppp=10;
				char nav[4096]=""; /* FIXME */
				char url[4096]=""; /* FIXME */
				
				npages=(Res->total_found/Agent->page_size)+1;
				
				page1=Agent->page_number-ppp/2;
				page2=Agent->page_number+ppp/2;
				if(page1<0){
					page2-=page1;
					page1=0;
				}else
				if(page2>npages){
					page1-=(page2-npages);
					page2=npages;
				}else{
				}
				if(page1<0)page1=page1=0;
				if(page2>npages)page2=npages;
				
				for(i=page1;i<page2;i++){
					UdmReplaceIntVar(Env->vars,"np",i,UDM_VARSRC_QSTRING);
					BuildPageURL(Env->vars,url);
					UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"NP",i+1,UDM_VARSRC_GLOBAL);
					PrintTemplate(Agent,NULL,UDM_STREND(nav),Env->vars,tmpl,(i==Agent->page_number)?T_NAVBAR0:T_NAVBAR1);
				}
				UdmAddStrVar(Env->vars,"NB",nav,UDM_VARSRC_GLOBAL);

				UdmReplaceIntVar(Env->vars,"np",Agent->page_number-1,UDM_VARSRC_QSTRING);
				BuildPageURL(Env->vars,url);
				UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);
				
				if(Res->first==1){/* First page */
					PrintTemplate(Agent,NULL,nav,Env->vars,tmpl,T_NAVLEFT_NOP);
					UdmReplaceStrVar(Env->vars,"NL",nav,UDM_VARSRC_GLOBAL);
				}else{
					PrintTemplate(Agent,NULL,nav,Env->vars,tmpl,T_NAVLEFT);
					UdmReplaceStrVar(Env->vars,"NL",nav,UDM_VARSRC_GLOBAL);
				}

				UdmReplaceIntVar(Env->vars,"np",Agent->page_number+1,UDM_VARSRC_QSTRING);
				BuildPageURL(Env->vars,url);
				UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);

				if(Res->last==Res->total_found){/* Last page */
					PrintTemplate(Agent,NULL,nav,Env->vars,tmpl,T_NAVRIGHT_NOP);
					UdmReplaceStrVar(Env->vars,"NR",nav,UDM_VARSRC_GLOBAL);
				}else{
					PrintTemplate(Agent,NULL,nav,Env->vars,tmpl,T_NAVRIGHT);
					UdmReplaceStrVar(Env->vars,"NR",nav,UDM_VARSRC_GLOBAL);
				}

				PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_RESTOP);
				for(i=0;i<Res->num_rows;i++){
					UDM_RESULT   * ClRes;
					UDM_DOCUMENT * Doc=&Res->Doc[i];
					UDM_ALIAS    * Alias;
					char clist[4096]=""; /* FIXME */
					char last_mod[128];
					char aliastr[UDM_URLSIZE];

					/* Build clone list */
					if((ClRes=UdmCloneList(Agent,Doc))){
						int c;
						
						for(c=0;c<ClRes->num_rows;c++){
							UDM_DOCUMENT * Clone=&ClRes->Doc[c];
							if(Res->Doc[i].url_id!=Clone->url_id){
								UdmReplaceIntVar(Env->vars,"DI",Clone->url_id,UDM_VARSRC_GLOBAL);
								UdmReplaceStrVar(Env->vars,"DU",Clone->url,UDM_VARSRC_GLOBAL);
								UdmReplaceStrVar(Env->vars,"DC",Clone->content_type,UDM_VARSRC_GLOBAL);
								UdmReplaceIntVar(Env->vars,"DS",Clone->size,UDM_VARSRC_GLOBAL);
								UdmTime_t2HttpStr(Clone->last_mod_time,last_mod);
								UdmReplaceStrVar(Env->vars,"DM",last_mod,UDM_VARSRC_GLOBAL);
								PrintTemplate(Agent,NULL,clist,Env->vars,tmpl,T_CLONE);
								
								/*sprintf(UDM_STREND(clist),"%s",Clone->url);*/
							}
						}
						UdmFreeResult(ClRes);
					}
					UdmReplaceStrVar(Env->vars,"CL",clist,UDM_VARSRC_GLOBAL);

					UdmReplaceIntVar(Env->vars,"DI",Doc->url_id,UDM_VARSRC_GLOBAL);
					if((Doc->url)&&(Alias=UdmAliasFind(&Agent->Conf->AliasList,Doc->url,0,NULL))){
						snprintf(aliastr,sizeof(aliastr),"%s%s",Alias->replace,Doc->url+strlen(Alias->match.pattern));
					}else{
						snprintf(aliastr,sizeof(aliastr),"%s",Doc->url?Doc->url:"");
					}

					UdmReplaceStrVar(Env->vars,"DU",aliastr,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DT",Doc->title,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DX",Doc->text,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DC",Doc->content_type,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DD",Doc->description,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DK",Doc->keywords,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DR",Doc->rating,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DS",Doc->size,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DN",Doc->order,UDM_VARSRC_GLOBAL);
					UdmTime_t2HttpStr(Doc->last_mod_time,last_mod);
					UdmReplaceStrVar(Env->vars,"DM",last_mod,UDM_VARSRC_GLOBAL);
					
					/* DE is DD or DT */
					if(Doc->description&&Doc->description[0])
						UdmReplaceStrVar(Env->vars,"DE",Doc->description,UDM_VARSRC_GLOBAL);
					else
						UdmReplaceStrVar(Env->vars,"DE",Doc->text,UDM_VARSRC_GLOBAL);
					
					PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_RESULT);
				}
				PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_RESBOT);
			}else{
				PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_NOTFOUND);
			}
			UdmFreeResult(Res);
		}else{
			if(UdmDBErrorCode(Agent->db)){
				UdmAddStrVar(Env->vars,"E",UdmDBErrorMsg(Agent->db),UDM_VARSRC_GLOBAL);
				PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_ERROR);
			}
		}
	}

	PrintTemplate(Agent,stdout,NULL,Env->vars,tmpl,T_BOTTOM);

	FreeTemplate(tmpl);
	UdmFreeAgent(Agent);
	UdmFreeEnv(Env);

	return(0);
} 
