#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_stopwords.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_agent.h"
#include "udm_unicode.h"


UDM_STOPWORD * UdmIsStopWord(UDM_ENV * Conf,char *word){
int low = 0;
int high = Conf->nstoplist - 1;
int middle, match;

	if(!Conf->stoplist)return(0);
	while (low <= high) {
		middle = (low + high) / 2;
		match = strcmp(Conf->stoplist[middle].word, word);
		if (match < 0)low = middle + 1;
		if (match > 0)high = middle - 1;
		if (match == 0)return(&Conf->stoplist[middle]);
	}
	return(NULL);
}

static int cmpstop(const void *s1,const void *s2){
	return(strcmp(((const UDM_STOPWORD*)s1)->word,((const UDM_STOPWORD*)s2)->word));
}

void UdmSortStopList(UDM_ENV * Conf){
	/* Sort stoplist to run binary search later */
	qsort((void*)Conf->stoplist,Conf->nstoplist,sizeof(UDM_STOPWORD),cmpstop);
}

int UdmAddStopWord(UDM_ENV * Conf,UDM_STOPWORD * stopword){
	size_t j;

	/* If the word is already in list     */
	/* We will not add it again           */
	/* But mark it as "international word"*/
	/* i.e. the word without language     */
	/* It will allow to avoid troubles    */
	/* with language guesser              */
	for(j=0;j<Conf->nstoplist;j++){
		if(!strcmp(Conf->stoplist[j].word,stopword->word)){
			UDM_FREE(Conf->stoplist[j].lang);
			Conf->stoplist[j].lang=strdup("");
			return 0;
		}
	}

	if (!Conf->nstoplist) {
		Conf->stoplist=(UDM_STOPWORD *)UdmXmalloc((Conf->nstoplist+1)*sizeof(UDM_STOPWORD));
	}else{
		Conf->stoplist=(UDM_STOPWORD *)UdmXrealloc(Conf->stoplist,(Conf->nstoplist+1)*sizeof(UDM_STOPWORD));
	}
	
	Conf->stoplist[Conf->nstoplist].word=strdup(stopword->word);
	Conf->stoplist[Conf->nstoplist].lang=strdup(stopword->lang?stopword->lang:"");
	Conf->nstoplist++;

	return(1);
}

void UdmFreeStopList(UDM_ENV * Env){
	size_t i;
	for(i=0;i<Env->nstoplist;i++){
		UDM_FREE(Env->stoplist[i].word);
		UDM_FREE(Env->stoplist[i].lang);
	}
	UDM_FREE(Env->stoplist);
}


__INDLIB__ int UdmFileLoadStopList(UDM_ENV * Conf,char * stoplist_file_name){
	char fname[1024];
	char str[1024];
	char * lasttok;
	FILE * stopfile;
	UDM_STOPWORD stopword;
	UDM_CHARSET * cs=NULL, * lcs=NULL;
	char * charset=NULL;

	if (!stoplist_file_name)
	{
		snprintf(fname,sizeof(fname)-1,"%s%c%s",UDM_CONF_DIR,UDMSLASH,"stopwords.txt");
	} else
	{
		strncpy(fname,stoplist_file_name,sizeof(fname)-1);
	}

	if (!(stopfile=fopen(fname,"r")))
	{
		sprintf(Conf->errstr,"Can't open stopwords file '%s' (%s)", fname, strerror(errno));
		Conf->errcode=1;
		return(1);
	}

	bzero(&stopword,sizeof(stopword));
	lcs=UdmGetCharSet(Conf->local_charset?Conf->local_charset:"utf-8");

	while(fgets(str,sizeof(str),stopfile)){
		if(!str[0])continue;
		if(str[0]=='#')continue;
		
		if(!strncmp(str,"Charset:",8)){
			UDM_FREE(charset);
			charset=UdmGetToken(str+8," \t\n\r",&lasttok);
			if(charset){
				charset=strdup(charset);
			}
		}else
		if(!strncmp(str,"Language:",9)){
			UDM_FREE(stopword.lang);
			stopword.lang=UdmGetToken(str+9," \t\n\r",&lasttok);
			if(stopword.lang)stopword.lang=strdup(stopword.lang);
		}else
		if((stopword.word=UdmGetToken(str,"\t\n\r",&lasttok))){
			int uwrd[UDM_MAXWORDSIZE+1], ures;
			
			if(!cs){
				if(!charset){
					sprintf(Conf->errstr,"No charset definition in stopwords file '%s'", fname);
					Conf->errcode=1;
					return(1);
				}else{
					cs=UdmGetCharSet(charset);
					if(!cs){
						sprintf(Conf->errstr,"Unknown charset '%s' in stopwords file '%s'", charset,fname);
						Conf->errcode=1;
						return(1);
					}
				}
			}
			ures=cs->to_uni(uwrd,stopword.word,sizeof(uwrd)-1);
			ures=lcs->from_uni(stopword.word,uwrd,UDM_MAXWORDSIZE-1);
			UdmAddStopWord(Conf,&stopword);
			/*fprintf(stderr,"'%s' '%s' '%s'\n",stopword.word,stopword.lang,charset);*/
		}
	}
	fclose(stopfile);
	UdmSortStopList(Conf);
	UDM_FREE(stopword.lang);
	return(0);
}
