#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_env.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_word.h"
#include "udm_synonym.h"


void UdmSynonymListInit(UDM_SYNONYMLIST * List){
	bzero(List,sizeof(*List));
}

int UdmSynonymListLoad(UDM_ENV * Env,const char * filename){
	FILE * f;
	char str[512];
	char lang[64]="";
	UDM_CHARSET * cs=NULL;

	if(!(f=fopen(filename,"r"))){
		Env->errcode=1;
		snprintf(Env->errstr,sizeof(Env->errstr)-1,"Can't open synonyms file '%s'",filename);
		return -1;
	}
	while(fgets(str,sizeof(str),f)){
		if(str[0]=='#'||str[0]==' '||str[0]=='\t'||str[0]=='\r'||str[0]=='\n')continue;
		
		if(!strncmp(str,"Charset:",8)){
			char * lasttok;
			char * charset;
			if((charset=UdmGetToken(str+8," \t\n\r",&lasttok))){
				cs=UdmGetCharSet(charset);
				if(!cs){
					snprintf(Env->errstr,sizeof(Env->errstr)-1,"Unknown charset '%s' in synonyms file '%s'",charset,filename);
					fclose(f);
					return(1);
				}
			}
		}else
		if(!strncmp(str,"Language:",9)){
			char * lasttok;
			char * l;
			if((l=UdmGetToken(str+9," \t\n\r",&lasttok))){
				strncpy(lang,l,sizeof(lang)-1);
			}
		}else{
			char *p,*s;
			char * lasttok;
			size_t plen;
			size_t slen;
			if(!cs){
				snprintf(Env->errstr,sizeof(Env->errstr)-1,"No Charset command in synonyms file '%s'",filename);
				fclose(f);
				return(1);
			}
			if(!lang[0]){
				snprintf(Env->errstr,sizeof(Env->errstr)-1,"No Language command in synonyms file '%s'",filename);
				fclose(f);
				return(1);
			}
			p=UdmGetToken(str," \t\r\n",&lasttok);
			if(!p)continue;
			s=UdmGetToken(NULL," \t\r\n",&lasttok);
			if(!s)continue;
			
			plen=strlen(p)+1;
			slen=strlen(s)+1;
			
			if((Env->SynList.nsynonyms+1)>=Env->SynList.msynonyms){
				Env->SynList.msynonyms+=64;
				Env->SynList.Synonyms=(UDM_SYNONYM*)realloc(Env->SynList.Synonyms,sizeof(UDM_SYNONYM)*Env->SynList.msynonyms);
			}
			
			bzero(&Env->SynList.Synonyms[Env->SynList.nsynonyms],sizeof(UDM_SYNONYM));
			
			/* Add direct order */
			Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword=(int*)malloc(plen*sizeof(int));
			cs->to_uni(Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword,p,plen);
			UdmUniStrToLower(Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword);
			
			Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword=(int*)malloc(slen*sizeof(int));
			cs->to_uni(Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword,s,slen);
			UdmUniStrToLower(Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword);

			Env->SynList.nsynonyms++;
			
			bzero(&Env->SynList.Synonyms[Env->SynList.nsynonyms],sizeof(UDM_SYNONYM));
			
			/* Add reverse order */
			Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword=(int*)malloc(slen*sizeof(int));
			cs->to_uni(Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword,s,slen);
			UdmUniStrToLower(Env->SynList.Synonyms[Env->SynList.nsynonyms].p.uword);
			
			Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword=(int*)malloc(plen*sizeof(int));
			cs->to_uni(Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword,p,plen);
			UdmUniStrToLower(Env->SynList.Synonyms[Env->SynList.nsynonyms].s.uword);

			Env->SynList.nsynonyms++;
		}
	}
	fclose(f);
	return 0;
}

void UdmSynonymListFree(UDM_SYNONYMLIST * List){
	size_t i;
	
	for(i=0;i<List->nsynonyms;i++){
		UDM_FREE(List->Synonyms[i].p.word);
		UDM_FREE(List->Synonyms[i].p.uword);
		UDM_FREE(List->Synonyms[i].s.word);
		UDM_FREE(List->Synonyms[i].s.uword);
	}
}

static int cmpsyn(const void * v1,const void * v2){
	const UDM_SYNONYM * s1=(const UDM_SYNONYM*)v1;
	const UDM_SYNONYM * s2=(const UDM_SYNONYM*)v2;
	return(UdmUniStrCmp(s1->p.uword,s2->p.uword));
}

void UdmSynonymListSort(UDM_SYNONYMLIST * List){
	qsort(List->Synonyms,List->nsynonyms,sizeof(UDM_SYNONYM),&cmpsyn);
}


UDM_WIDEWORDLIST * UdmSynonymListFind(const UDM_SYNONYMLIST * List,UDM_WIDEWORD * wword){
	UDM_SYNONYM syn,*res,*first,*last;
	UDM_WIDEWORDLIST * Res=NULL;

	syn.p.uword=wword->uword;

	res=bsearch(&syn,List->Synonyms,List->nsynonyms,sizeof(UDM_SYNONYM),&cmpsyn);

	if(res){
		size_t nnorm,i;

		Res=(UDM_WIDEWORDLIST *)malloc(sizeof(*Res));
		UdmWideWordListInit(Res);

		/* Find first and last synonym */
		for(first=res;first>List->Synonyms;first--){
			if(UdmUniStrCmp(wword->uword,first->p.uword)){
				break;
			}else{
				first->s.order=wword->order;
				UdmWideWordListAdd(Res,&first->s);
			}
		}
		for(last=res+1;last<List->Synonyms+List->nsynonyms;last++){
			if(UdmUniStrCmp(wword->uword,last->p.uword)){
				break;
			}else{
				last->s.order=wword->order;
				UdmWideWordListAdd(Res,&last->s);
			}
		}

		/* Now find each of them in reverse order */
		nnorm=Res->nwords;
		for(i=0;i<nnorm;i++){

			syn.p.uword=Res->Words[i].uword;
			res=bsearch(&syn,List->Synonyms,List->nsynonyms,sizeof(UDM_SYNONYM),&cmpsyn);
			
			if(res){
				/* Find first and last synonym */
				for(first=res;first>List->Synonyms;first--){
					if(UdmUniStrCmp(syn.p.uword,first->p.uword)){
						break;
					}else{
						first->s.order=wword->order;
						UdmWideWordListAdd(Res,&first->s);
					}
				}
				for(last=res+1;last<List->Synonyms+List->nsynonyms;last++){
					if(UdmUniStrCmp(syn.p.uword,last->p.uword)){
						break;
					}else{
						last->s.order=wword->order;
						UdmWideWordListAdd(Res,&last->s);
					}
				}
			}
		}
	}
	return(Res);
}
