#include <udm_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_agent.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_services.h"
#include "udm_hrefs.h"

#define UDM_URLD_MAXCLNTS 128

#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long)-1)
#endif

static char pidname[1024];
static UDM_ENV * Conf=NULL;

static void exitproc(void){
	unlink(pidname);
}

static char time_pid[100];
static char * time_pid_info(void){
	struct tm * tim;
	time_t t;
	t=time(NULL);
	tim=localtime(&t);
	strftime(time_pid,sizeof(time_pid),"%a %d %T",tim);
	sprintf(time_pid+strlen(time_pid)," [%d]",(int)getpid());
	return(time_pid);
}

/********************* SIG Handlers ****************/
static void sighandler(int sign);
static void init_signals(void){
	/* Set up signals handler*/
	signal(SIGPIPE,sighandler);
	signal(SIGHUP,sighandler);
	signal(SIGINT,sighandler);
	signal(SIGTERM,sighandler);
}
static void sighandler(int sign){

	fprintf(stderr,"%s Signal %d arrived\n",time_pid_info(),sign);
	switch(sign){
		case SIGPIPE:
			fprintf(stderr,"%s FATAL: This should not ever happen! Broken pipe!\n",time_pid_info());
			break;
		case SIGHUP:
			break;
		default:
			fprintf(stderr,"%s Shutdown\n",time_pid_info());
			exit(0);
	}
	init_signals();
}

/*************************************************************/

typedef struct {
	int id;
	int fd;
	size_t nbytes;
	size_t mbytes;
	char * buf;
} UDM_URLD_CL;

static void init_client(UDM_URLD_CL *cl, int n){
	int i;
	
	for(i=0;i<n;i++){
		cl[i].id=0;
		cl[i].fd=0;
		cl[i].nbytes=0;
		cl[i].mbytes=1024;
		cl[i].buf=(char*)malloc(cl[i].mbytes);
		bzero(cl[i].buf,cl[i].mbytes);
	}
}

static void close_client(UDM_URLD_CL * cl, int n){
	int i;
	
	for(i=0;i<n;i++){
		close(cl[i].fd);
		UDM_FREE(cl[i].buf);
	}
}

#define XVAL(x)	((x)?(x):"-")
static int send_doc(UDM_URLD_CL * cl,UDM_AGENT * Indexer){
	char buf[1024*5];
	UDM_DOCUMENT * Doc;
	int index_flags=0; /* FIXME */
	int result, sent;
	size_t len;

	if(UdmStoreHrefs(Indexer))Indexer->Conf->have_targets=1;
	if(!(Doc=UdmGetDocInfo(Indexer,index_flags))){ 
		if(UdmDBErrorCode(Indexer->db)){ 
			result=IND_ERROR;
			fprintf(stderr,"Error: '%s'",UdmAgentErrorMsg(Indexer));
		}else{ 
			Indexer->Conf->have_targets=0; 
			result=IND_NO_TARGET;
		}
		sprintf(buf,"-\n");
	}else{ 
		sprintf(buf,"%s\t%li\t%d\t%d\t%d\t%d\t%d\n",XVAL(Doc->url),Doc->last_mod_time,Doc->url_id,Doc->size,Doc->status,Doc->hops,Doc->crc32);
	}
	
	len=strlen(buf);
	sent=send(cl->fd, (void *)&buf, len,0);
	printf("[%d] send cmd select '%s' %d/%d\n",cl->id,buf,sent,len);
	return 0;
}

static int send_origin(UDM_URLD_CL * cl,int origin){
	char buf[64];
	sprintf(buf,"%d\n",origin);
	return(send(cl->fd, (void *)&buf, strlen(buf),0));
}


#define UDM_URLD_CMD_SELECT	1
#define UDM_URLD_CMD_LUPDATE	2
#define UDM_URLD_CMD_CLONE	3
#define UDM_URLD_CMD_INSERT	4
#define UDM_URLD_CMD_SUPDATE	5

struct {
	int cmd;
	const char * strcmd;
} cmd_arr[] =
{	
	{UDM_URLD_CMD_SELECT,"select"},
	{UDM_URLD_CMD_LUPDATE,"lupdate"},
	{UDM_URLD_CMD_SUPDATE,"supdate"},
	{UDM_URLD_CMD_INSERT,"insert"},
	{UDM_URLD_CMD_CLONE, "clone"},
	{0, ""}
};

static int parse_cmd(UDM_URLD_CL * cl, UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	char * tok, * lt;
	int ntok=0,res=0, origin=0;
	
	tok=UdmGetToken(cl->buf,"\t\r\n",&lt);
	if(tok){
		int i;
		
		for(i=0;cmd_arr[i].cmd>0;i++){
			if(!strcmp(cmd_arr[i].strcmd,tok)){
				res=cmd_arr[i].cmd;
				break;
			}
		}
		printf("[%d] recv cmd %s\n",cl->id,tok);
	}
	switch(res){
		case UDM_URLD_CMD_SELECT:
			send_doc(cl,Indexer);
			break;
		case UDM_URLD_CMD_INSERT:
			while(tok){
				switch(ntok){
					case 1: Doc->url=tok;break;
					case 2: Doc->referrer=atoi(tok);break;
					case 3: Doc->hops=atoi(tok);break;
					case 4: Doc->tag=tok;break;
					case 5: Doc->category=tok;break;
				}
				ntok++;
				tok=UdmGetToken(NULL,"\t\r\n",&lt);
			}
			if(Doc->url){
				if(UdmAddHref(&Indexer->Conf->HrefList,Doc))
					Indexer->Conf->have_targets=1;
			}
			break;
		case UDM_URLD_CMD_CLONE:
			while(tok){
				switch(ntok){
					case 1: Doc->url_id=atoi(tok);break;
					case 2: Doc->crc32=atoi(tok);break;
					case 3: Doc->size=atoi(tok);break;
				}
				ntok++;
				tok=UdmGetToken(NULL,"\t\r\n",&lt);
			}
			origin=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_FINDORIG);
			origin=((origin==Doc->url_id)?0:origin);
			send_origin(cl,origin);
			break;
		case UDM_URLD_CMD_SUPDATE:
			while(tok){
				switch(ntok){
					case 1: Doc->url_id=atoi(tok);break;
					case 2: Doc->status=atoi(tok);break;
					case 3: Doc->next_index_time=atoi(tok);break;
				}
				ntok++;
				tok=UdmGetToken(NULL,"\t\r\n",&lt);
			}
			res=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_SUPDATE);
			break;

		case UDM_URLD_CMD_LUPDATE:
			while(tok){
				switch(ntok){
					case 1:  Doc->url_id=atoi(tok);break;
					case 2:  Doc->status=atoi(tok);break;
					case 3:  Doc->size=atoi(tok);break;
					case 4:  Doc->referrer=atoi(tok);break;
					case 5:  Doc->hops=atoi(tok);break;
					case 6:  Doc->crc32=atoi(tok);break;
					case 7:  Doc->last_mod_time=atoi(tok);break;
					case 8:  Doc->next_index_time=atoi(tok);break;
					case 9:  Doc->tag=tok;break;
					case 10: Doc->category=tok;break;
					case 11: Doc->url=tok;break;
					case 12: Doc->content_type=tok;break;
					case 13: Doc->title=tok;break;
					case 14: Doc->keywords=tok;break;
					case 15: Doc->description=tok;break;
					case 16: Doc->text=tok;break;
				}
				ntok++;
				tok=UdmGetToken(NULL,"\t\r\n",&lt);
			}
			res=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_LUPDATE);
			break;
	}
	return res;
}

static void do_client(UDM_URLD_CL * cl, UDM_AGENT * Indexer){
	UDM_DOCUMENT Doc;
	
	bzero(&Doc,sizeof(Doc));
	parse_cmd(cl,Indexer,&Doc);
}

int main(int argc, char **argv) {
	struct sockaddr_in server_addr, client_addr;
	int ctl_sock, nport, on=1, id=1;
	int pid_fd;
	char pidbuf[1024];
	const char *config_name= UDM_CONF_DIR "/urld.conf";
	UDM_URLD_CL cl[UDM_URLD_MAXCLNTS];
	UDM_AGENT * Indexer;
	fd_set mask;

	/* Check that another instance isn't running */
	/* and create PID file.                      */

	sprintf(pidname,"%s/%s",UDM_VAR_DIR,"urld.pid");
	pid_fd = open(pidname,O_CREAT|O_EXCL|O_WRONLY,0644);
	if(pid_fd < 0){
		fprintf(stderr,"%s Can't create '%s': %s\n", time_pid_info(), pidname, strerror(errno));
		if(errno == EEXIST){
			fprintf(stderr,"It seems that another urld is already running!\n");
			fprintf(stderr,"Remove '%s' if it is not true.\n",pidname);
 		}
		exit(1);
	}
	
	sprintf(pidbuf,"%d\n",(int)getpid());
	write(pid_fd,&pidbuf,strlen(pidbuf));

	atexit(&exitproc);
	init_signals();

	UdmInit(); /* Initialize library */
	init_client(cl,UDM_URLD_MAXCLNTS);
	Conf = UdmAllocEnv();

	UdmLoadConfig(Conf,config_name,0,0);

	if(UdmEnvErrCode(Conf)){
 		fprintf(stderr,"%s\n",UdmEnvErrMsg(Conf));
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}

	Indexer=UdmAllocAgent(Conf,0,UDM_OPEN_MODE_WRITE);

	nport = UDM_URLD_PORT;
	nport = htons((u_short) nport);

	if ((ctl_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr,"%s socket() error %d\n",time_pid_info(),errno);
		unlink(pidname);
		exit(1);
	}
	if (setsockopt(ctl_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) != 0){
		fprintf(stderr,"%s setsockopt() error %d\n",time_pid_info(),errno);
		unlink(pidname);
		exit(1);
	}

	bzero(&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = nport;
	

	if (bind(ctl_sock, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
		fprintf(stderr,"%s bind() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		exit(1);
	}

	if (listen(ctl_sock, UDM_URLD_MAXCLNTS) == -1) {
		fprintf(stderr,"%s listen() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		exit(1);
	}
	
	FD_ZERO(&mask);
	FD_SET(ctl_sock,&mask);

	while(1) {
		struct timeval tval;
		int ns, sel, i;
		fd_set msk;
		
		tval.tv_sec=300;
		tval.tv_usec=0;
		msk=mask;
		sel=select(UDM_URLD_MAXCLNTS,&msk,0,0,&tval);
		
		printf("select=%d\n",sel);
		if(sel==0)continue;
		if(sel==-1){
			printf("error %d %s\n",errno,strerror(errno));
			continue; /* FIXME */
		}
		
		if(FD_ISSET(ctl_sock,&msk)){
			socklen_t addrlen;
			
			for(i=0;i<UDM_URLD_MAXCLNTS;i++){
				if(cl[i].fd==0){
					bzero(&client_addr, addrlen = sizeof(client_addr));
					if ((ns = accept(ctl_sock, (struct sockaddr *) &client_addr, &addrlen)) == -1) {
						fprintf(stderr,"%s accept() error %d %s\n",time_pid_info(),errno,strerror(errno));
						unlink(pidname);
						exit(1);
					}
					FD_SET(ns,&mask);
					cl[i].id=id++;
					cl[i].fd=ns;
					fprintf(stderr, "Client = %s\n", inet_ntoa(client_addr.sin_addr));
					break;
				}
			}
		}

		for(i=0;i<UDM_URLD_MAXCLNTS;i++){
			if(FD_ISSET(cl[i].fd,&msk)){
				size_t lbytes,rbytes;
				
				rbytes=cl[i].mbytes-cl[i].nbytes-1;
				lbytes=recv(cl[i].fd,cl[i].buf+cl[i].nbytes,rbytes,0);
				
				printf("[%d] recv\n",cl[i].id);fflush(stdout);
				write(1,cl[i].buf+cl[i].nbytes,lbytes);
				printf("\n");fflush(stdout);
				
				if(lbytes){
					char * eoc;
					
					cl[i].nbytes+=lbytes;
					cl[i].buf[cl[i].nbytes]='\0';
					
					while((eoc=strchr(cl[i].buf,'\n'))){
						size_t remain;
						
						*eoc='\0';
						remain=cl[i].nbytes-(eoc-cl[i].buf+1);
						/*printf("BUF='%s'\n",cl[i].buf);*/
						do_client(&cl[i],Indexer);
						if(remain){
							memmove(cl[i].buf,eoc+1,remain+1);
						}
						cl[i].nbytes=remain;
					}
				}else{
					fprintf(stderr,"%s Client #%d left\n",time_pid_info(),i);
					FD_CLR(cl[i].fd,&mask);
					close_client(&cl[i],1);
					init_client(&cl[i],1);
				}
			}
		}
	}
}
